import debounce from "debounce";
import util from "util";

export default function funCache<T extends Function>(func: T, options: FCOptions = { lifeTime: 0, debounceTimer: 1000, async: false }): T & { clearCache: () => void } {
  let cached: any = {
    ____timeOfCreation: Date.now(),
    ...options.initialCache
  };

  const updateData = options.onDataUpdate ? debounce(() => {
    try {
      options.onDataUpdate?.(cached);
    } catch (err) {
      if (options.debug) {
        console.error(err)
      }
    }
  }, options.debounceTimer) : undefined;

  const checkExpiry = () => {
    if (options.lifeTime !== 0 && Date.now() - cached.____timeOfCreation > options.lifeTime)
      cached = { ____timeOfCreation: cached.____timeOfCreation }
  }

  const cache = (getnewval: Function, str: any) => {
    checkExpiry();


    if (str in cached) return cached[str];
    cached[str] = getnewval();

    updateData?.();

    return cached[str]
  }

  const asyncCache = async (getnewval: Function, str: any) => {
    checkExpiry();

    if (str in cached) return cached[str];
    cached[str] = await getnewval();

    updateData?.();

    return cached[str]
  }

  if (options.async || util.types.isAsyncFunction(func)) {
    return (async (...args: any) =>
      await asyncCache(() => func(...args), args.join(","))
    ) as any
  }

  const cachedFN = ((...args: any) =>
    cache(func(...args), args.join(","))
  ) as any;

  cachedFN.clearCache = () => {
    cached = { ____timeOfCreation: Date.now() }
  }

  cachedFN.noCache = (...args: any) => func(...args);

  return cachedFN;
}



export function localStorageCacher(tmpPath: string, options?: { localStorage?: Storage }) {
  const ls = options?.localStorage || window.localStorage;
  return {
    initialCache: JSON.parse(
      ls.getItem(tmpPath) || "{}"
    ),
    onDataUpdate: (ndata: any) => {
      ls.setItem(tmpPath, JSON.stringify(ndata))
    },
  };
}




export function fSCacher(tmpPath: string) {
  const fs = require("fs");
  const fsAsync = require("fs/promises")
  return {
    initialCache: JSON.parse(
      // @ts-ignore
      fsAsync.existsSync(tmpPath)
        ? fsAsync.readFileSync(tmpPath, {
          encoding: "utf-8",
        })
        : "{}"
    ),
    onDataUpdate: async (ndata: any) => {
      try {
        await fs.unlink(tmpPath);
        // eslint-disable-next-line no-empty
      } catch { }

      await fs.writeFile(tmpPath, JSON.stringify(ndata), {
        encoding: "utf-8",
        flag: "w",
      });
    },
  };
}

import type { RedisClientType } from 'redis';

export function redisCacher(tmpPath: string, options: { client: RedisClientType }): Partial<FCOptions> {
  const red = options.client;
  return {
    async getCache() {
      return JSON.parse(
        await red.get(tmpPath) || "{}"
      )
    },
    onDataUpdate: async (ndata: any) => {
      await red.set(tmpPath, JSON.stringify(ndata))
    },
  };
}

import type { Redis } from '@upstash/redis';

export function upstashCacher(tmpPath: string, options: { client: Redis }): Partial<FCOptions> {
  const red = options.client;
  return {
    async getCache() {
      return JSON.parse(
        await red.get(tmpPath) || "{}"
      )
    },
    onDataUpdate: async (ndata: any) => {
      await red.set(tmpPath, JSON.stringify(ndata))
    },
  };
}